﻿/*
This source file is part of the RGBZ Video Camera project published in the paper

  Coherent Spatiotemporal Filtering, Upsampling and Rendering of RGBZ Videos
  Christian Richardt, Carsten Stoll, Neil A. Dodgson, Hans-Peter Seidel, Christian Theobalt
  Computer Graphics Forum (Proceedings of Eurographics), 31(2), May 2012
  <http://www.mpi-inf.mpg.de/resources/rgbz-camera/>

If you use our code, please cite our paper. Thanks.

Copyright (c) 2012 Christian Richardt <christian@richardt.name>


Code licence: GPL v3
----------------------------------------------------------------------------------------------------
This program is free software; you can redistribute it and/or modify it under the terms
of the GNU General Public License as published by the Free Software Foundation;
either version 3 of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along with this program;
if not, see <http://www.gnu.org/licenses/>.

$Id: MultiResolutionFillIn.cs 537 2012-04-04 15:35:42Z cr333 $
*/

// uncomment to create intermediate image for multi-resolution fill-in figure in paper
//#define SAVE_INTERMEDIATE_IMAGES

using System.Collections.Generic;
using OpenTK.Graphics.OpenGL;
using VideoLib;
using VideoLib.Frames;
using VideoLib.OpenGL;
using System.ComponentModel;

namespace RGBZcamera.Library.FillIn
{
	/// <summary>Implements a multi-resolution fill-in procedure based on a joint-bilateral filter.</summary>
	public class MultiResolutionFillIn : Node, IOpenGLNode
	{
		/// <summary>Gets or sets the downsampling factor.</summary>
		[DefaultValue(2), Description("Downsampling factor between levels.")]
		public int Factor { get; set; }

		/// <summary>Gets or sets the number of levels.</summary>
		[DefaultValue(3), Description("Number of levels used in fill-in procedure.")]
		public int Levels { get; set; }

		/// <summary>Gets or sets the spatial sigma.</summary>
		[DefaultValue(10.0f), Description("Spatial extent of the joint-bilateral fill-in.")]
		public float SpatialSigma { get; set; }

		/// <summary>Gets or sets the range sigma.</summary>
		[DefaultValue(0.05f), Description("The range sigma describes how similar colours should be.")]
		public float RangeSigma { get; set; }

		/// <summary>Gets or sets whether points filled in lower levels are recomputed in higher levels.</summary>
		[DefaultValue(true), Description("Recompute the sparsely-upsampled points to avoid artefacts.")]
		public bool RecomputePoints { get; set; }

		/// <summary>Gets or sets a value indicating whether this <see cref="MultiResolutionFillIn"/> node is enabled.</summary>
		[DefaultValue(true), Description("Enables the multi-res fill-in procedure.")]
		public bool Enabled { get; set; }

		private FillInDownsamplingFilter   downsampler;
		private FillInJointBilateralFilter bilfilter;
		private FillInUpsamplingFilter     upsampler;

		/// <summary>Initializes a new instance of the <see cref="MultiResolutionFillIn"/> class.</summary>
		public MultiResolutionFillIn()
		{
			// default parameters
			Factor          = 2;
			Levels          = 3;
			SpatialSigma    = 10;
			RangeSigma      = 0.05f;
			RecomputePoints = true;
			Enabled         = true;

			// create nodes outside the current topology (parameters are set later)
			downsampler = new FillInDownsamplingFilter()   { Topology = null };
			bilfilter   = new FillInJointBilateralFilter() { Topology = null };
			upsampler   = new FillInUpsamplingFilter()     { Topology = null };

			// IO pins
			AddInputPins("colour", "points");
			AddOutputPin("out");
		}

		/// <summary>Initialisation code, executed just before the first StatusUpdate.</summary>
		public override void Init()
		{
			downsampler.Init();
			bilfilter.Init();
			upsampler.Init();
		}

		/// <summary>Processes some frames, to produce new frames.</summary>
		public override void Process()
		{
			// get inputs
			var frames = GetInputFrames();
			if (PassOnEndOfStreamFrame(frames)) return;
			var textures = CastFramesTo<TextureFrame>(frames);
			AssertFrameProperties(textures, FrameProps.Size);

			if (Enabled)
			{
				// set parameters
				downsampler.Factor        = Factor;
				upsampler.Factor          = Factor;
				bilfilter.SpatialSigma    = SpatialSigma;
				bilfilter.RangeSigma      = RangeSigma;
				bilfilter.RecomputePoints = RecomputePoints;

				// apply multi-resolution fill-in (called on top layer, #0)
				var result = RunLevel(textures["colour"], textures["points"], 0);

				// output result
				PushFrame(result);
				result.RefDec();
			}
			else // disabled: just return input points
			{
				PushFrame(textures["points"]);
				textures.RefDec();
			}
		}

		/// <summary>Runs a level of the multi-resolution joint-bilateral fill-in procedure.</summary>
		/// <param name="colour">The input colour texture.</param>
		/// <param name="points">The input points texture.</param>
		/// <param name="level">The number of the level to compute.</param>
		/// <returns>Points texture with filled points.</returns>
		private TextureFrame RunLevel(TextureFrame colour, TextureFrame points, int level)
		{
#if SAVE_INTERMEDIATE_IMAGES
			// save input colour image to this level
			var c_k = colour.Texture.ReadFrame(VideoLib.Frames.PixelFormat.BGR24);
			c_k.Bitmap.Save(string.Format("FillIn-C{0}.png", level));

			// save input distance map to this level
			var p_k = ReadDistanceMapFrame(points);
			p_k.Bitmap.Save(string.Format("FillIn-P{0}.png", level));
#endif

			// Levels in multi-resolution fill-in:
			// -----------------------------------
			// 0: highest resolution
			// 1: downsampled once
			// ...
			// n: coarsest resolution (n = Levels - 1)

			if (level == Levels - 1) // lowest layer
			{
				// Structure of lowest layer:
				// --------------------------
				//         ... from/to upper level ...
				//    |            ^                      |
				//    v            |                      v
				// [colour] --> [filled] <------------ [points]

				// just joint-bilateral fill-in
				points.RefInc(); // increase reference count, as FillIn will decrease it twice
				var filled = FillIn(colour, points, points);

#if SAVE_INTERMEDIATE_IMAGES
				// save filled distance map
				var j_k = ReadDistanceMapFrame(filled);
				j_k.Bitmap.Save(string.Format("FillIn-J{0}.png", level));
#endif

				return filled;
			}
			else // all other layers
			{
				// Structure of all but the lowest level:
				// --------------------------------------
				//         ... from/to upper level ...
				//    |            ^                      |
				//    v            |                      v
				// [colour] --> [filled] filter    +-- [points]
				//    |            ^               |      |
				//    |            |   upsample    |      |
				//    | down       +-- [sparse] <--+      | down
				//    | sample            ^               | sample
				//    |            +------+ filter        |
				//    v            |                      v
				//         ... to/from lower level ...

				// first downsample inputs
				var loResPoints = downsampler.Draw(new Dictionary<string, TextureFrame> { { "in", points } });
				var loResColour = downsampler.Draw(new Dictionary<string, TextureFrame> { { "in", colour } });

				// then run the lower layer recursively, which returns a filled-in points image of the same resolution as its inputs
				var loResJbfPoints = RunLevel(loResColour, loResPoints, level + 1);

				// so upsample the low-resolution version to the current resolution, which will then have sparsely filled-in values in a grid
				var hiResSparsePoints = upsampler.Draw(new Dictionary<string, TextureFrame> { { "lores", loResJbfPoints }, { "hires", points } });
				loResJbfPoints.RefDec();

#if SAVE_INTERMEDIATE_IMAGES
				// save filled distance map
				var s_k = ReadDistanceMapFrame(hiResSparsePoints);
				s_k.Bitmap.Save(string.Format("FillIn-S{0}.png", level));
#endif

				// apply the joint-bilateral fill-in
				var filled = FillIn(colour, hiResSparsePoints, points);

#if SAVE_INTERMEDIATE_IMAGES
				// save filled distance map
				var j_k = ReadDistanceMapFrame(filled);
				j_k.Bitmap.Save(string.Format("FillIn-J{0}.png", level));
#endif

				return filled;
			}
		}

		/// <summary>Fills invalid points using a joint-bilateral filter on the colour image.</summary>
		/// <param name="colour">The colour texture.</param>
		/// <param name="points">The points texture.</param>
		/// <param name="originalPoints">The original points texture, to know which points to fill in.</param>
		/// <returns>Points texture with filled points.</returns>
		private TextureFrame FillIn(TextureFrame colour, TextureFrame points, TextureFrame originalPoints)
		{
			// apply joint-bilateral fill-in
			var result = bilfilter.Draw(new Dictionary<string, TextureFrame> { { "colour", colour }, { "points", points }, { "rawpoints", originalPoints } });

			// cleanup
			colour.RefDec();
			points.RefDec();
			originalPoints.RefDec();

			return result;
		}

		/// <summary>Read back the distance map from a poitns texture (alpha component).</summary>
		/// <param name="points">Point cloud texture.</param>
		private static BitmapFrame ReadDistanceMapFrame(TextureFrame points)
		{
			points.Bind();
			VideoLib.Frames.BitmapFrame frame = new VideoLib.Frames.BitmapFrame(points.Width, points.Height, VideoLib.Frames.PixelFormat.Grey8);
			GL.PixelStore(PixelStoreParameter.PackAlignment, 4);
			GL.GetTexImage(TextureTarget.Texture2D, 0, OpenTK.Graphics.OpenGL.PixelFormat.Alpha, PixelType.UnsignedByte, frame.Data);
			points.Unbind();
			return frame;
		}
	}
}
