﻿/*
This source file is part of the RGBZ Video Camera project published in the paper

  Coherent Spatiotemporal Filtering, Upsampling and Rendering of RGBZ Videos
  Christian Richardt, Carsten Stoll, Neil A. Dodgson, Hans-Peter Seidel, Christian Theobalt
  Computer Graphics Forum (Proceedings of Eurographics), 31(2), May 2012
  <http://www.mpi-inf.mpg.de/resources/rgbz-camera/>

If you use our code, please cite our paper. Thanks.

Copyright (c) 2012 Christian Richardt <christian@richardt.name>


Code licence: GPL v3
----------------------------------------------------------------------------------------------------
This program is free software; you can redistribute it and/or modify it under the terms
of the GNU General Public License as published by the Free Software Foundation;
either version 3 of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along with this program;
if not, see <http://www.gnu.org/licenses/>.

$Id: FillInJointBilateralFilter.frag 537 2012-04-04 15:35:42Z cr333 $
*/
#extension GL_EXT_gpu_shader4 : enable

// input texture
uniform sampler2D colour;
uniform sampler2D points;
uniform sampler2D rawpoints;

// parameters
uniform float spatialSigma;
uniform float rangeSigma;
uniform bool recomputePoints;

// debug stuff
uniform bool onOff;
uniform bool debug;
uniform bool overlayColour;
uniform bool overlayWeights;
uniform vec2 weightPoint;
uniform float weightScale;
uniform bool threshMap;

void main()
{
	// look up texture size once and reuse it later,
	// assuming every texture has the same size
	vec2 size = vec2(textureSize2D(colour, 0));

	// precompute 1 / sigma^2
	float iSigmaS = 1.0 / (2.0 * spatialSigma * spatialSigma);
	float iSigmaR = 1.0 / (2.0 * rangeSigma   * rangeSigma);

	// truncate Gaussian after 2 sigma
	int radius = int(ceil(2.0 * spatialSigma));

	// colour and coordinate at the central pixel
	vec3 centreColour = texture2D(colour, gl_FragCoord.xy / size).rgb;
	vec4 centrePoint  = texture2D(points, gl_FragCoord.xy / size);
	
	// determine point validity
	float validPoint = centrePoint.w;
	if(recomputePoints)
		validPoint = texture2D(rawpoints, gl_FragCoord.xy / size).w;
	
	// debug: use thresholded distance map as colour image
	if(threshMap)
		centreColour = vec3(length(centrePoint.xyz) > 0.55 ? 1.0 : 0.0);
	
	// default result: unfiltered point
	gl_FragColor = centrePoint;
		
	// only fill in invalid pixels
	if(onOff && validPoint == 0.0)
	{	
		// homogeneous accumulator of depth, i.e. (wz, w)
		vec2 acc = vec2(0.0);

		for(int dy = -radius; dy <= radius; dy++)
		{
			for(int dx = -radius; dx <= radius; dx++)
			{
				vec2 coords = (gl_FragCoord.xy + vec2(dx, dy)) / size;

				// colour values at the current pixel
				vec3 currentColour = texture2D(colour, coords).rgb;
				vec4 currentPoint  = texture2D(points, coords);
				
				// debug: use thresholded distance map as colour image
				if(threshMap)
					currentColour = vec3(length(currentPoint.xyz) > 0.55 ? 1.0 : 0.0);

				// only accumulate valid points
				if(currentPoint.w != 0.0)
				{
					// compute pixel weight, using exp(a) * exp(b) = exp(a + b);
					float w = 0.0;

					// spatial weight
					vec2 deltaS = vec2(dx, dy);
					w += iSigmaS * dot(deltaS, deltaS);
					
					// range weight
					vec3 deltaR = currentColour - centreColour;
					w += iSigmaR * dot(deltaR, deltaR);

					acc += exp(-w) * vec2(currentPoint.w, 1.0);
				}
			}
		}

		//// filtered colour (black if there were no valid data samples in the kernel)
		//if(acc.y <= 0.0) gl_FragColor = centrePoint; // unfiltered
		//else             gl_FragColor = vec4(centrePoint.xyz, length(centrePoint.xyz)) / centrePoint.zzzz * (acc.xxxx / acc.yyyy);
		
		if(acc.y > 0.0)
		{
			//vec3 newPoint = centrePoint.xyz / centrePoint.zzz * (acc.xxx / acc.yyy);
			vec3 newPoint = centrePoint.xyz / length(centrePoint.xyz) * (acc.xxx / acc.yyy);
			gl_FragColor = vec4(newPoint, length(newPoint));
		}
	}
	
	//gl_FragColor = vec4(gl_FragColor.www, 1);
	
	// overlay the colour image
	if(overlayColour) gl_FragColor = mix(gl_FragColor, vec4(centreColour.rgb, 1), 0.5);
	
	// show kernel weight for one pixel
	if(overlayWeights)
	{
		vec3 pickColour = texture2D(colour, (weightPoint + vec2(0.5)) / size).rgb;
		
		if(threshMap)
		{
			vec3 pickPoint = texture2D(points, (weightPoint + vec2(0.5)) / size).rgb;
			pickColour = vec3(length(pickPoint.xyz) > 0.55 ? 1.0 : 0.0);
		}
		
		float w = 0.0;
		
		// spatial weight
		vec2 deltaS = (weightPoint + vec2(0.5)) - gl_FragCoord.xy;
		
		// range weight
		vec3 deltaR = pickColour - centreColour;

		vec4 overlay = vec4(0.0);
		if(centrePoint.w > 0.0)
			overlay = weightScale * vec4(exp(-iSigmaS * dot(deltaS, deltaS) - iSigmaR * dot(deltaR, deltaR)));

		// mark point as red
		if(int(gl_FragCoord.x) == int(weightPoint.x) && int(gl_FragCoord.y) == int(weightPoint.y))
			overlay = vec4(1.0, 0.0, 0.0, 1.0);
			
		gl_FragColor = mix(gl_FragColor, overlay, 0.5);
	}
	
	if(debug)
	{
		if(centrePoint.w == 0.0) gl_FragColor += 0.2;
		//if(centrePoint.w == 0.0) gl_FragColor = mix(gl_FragColor, vec4(1.0), 0.2);
		//else                     gl_FragColor = mix(gl_FragColor, vec4(0.0), 0.2);
	}
}