// LibAvWrap -- .NET wrapper for the libav* libraries
// --------------------------------------------------
// By Christian Richardt <christian@richardt.name>
// (based on AvDn 0.7, www.informikon.com/tutorials/)
//
// $Id: Utils.h 436 2010-10-15 16:26:38Z cr333 $

#pragma once

using namespace System;
using namespace System::Text;
using namespace System::Runtime::InteropServices;
using namespace System::Threading;

extern "C" {
#include "libavutil/log.h"
#include <stdio.h>
#include <string.h>
}

#include "Enums.h"

// printf-based logger for av_log.
inline void printf_callback(void* ptr, int level, const char* fmt, va_list vl)
{
	static int print_prefix = 1;
	if(level > av_log_get_level()) return;

	AVClass* avc = ptr ? *(AVClass**)ptr : NULL;
	if(print_prefix && avc)
	{
		String^ threadName = Thread::CurrentThread->Name;
		IntPtr ipThreadName = Marshal::StringToHGlobalAnsi(threadName);
		char* strThreadName = (char*)ipThreadName.ToPointer();

		// [01:12345678:Thread1:swscaler] LOG ENTRY
		fprintf(stdout, "[%i:%p:%s:%-8s] ", level, avc, strThreadName, avc->item_name(ptr));

		Marshal::FreeHGlobal(ipThreadName);
	}

	print_prefix = (strstr(fmt, "\n") != NULL);
	vfprintf(stdout, fmt, vl);
}

///<summary>Wrapper for some lavc and lavf structures.</summary>
namespace LibAvWrap
{
	// forward declaration(s)
	ref class LibAvByteIOContext;
	ref class LibAvFormatContext;
	ref class LibAvOutputFormat;

	///<summary>A general utility class.</summary>
	public ref class LibAvUtils
	{

	public:
// constructors and destructor --------------------------------------------------------------------

// member functions -------------------------------------------------------------------------------

		///<summary>Free memory allocated by av_malloc.</summary>
		///<remarks>Calls av_free.</remarks>
		static void AvFree(IntPtr ptr)
		{
			av_free((void*)ptr);
		}

		///<summary>Finds the format corresponding to the short name.</summary>
		///<remarks>Calls guess_format.</remarks>
		static LibAvOutputFormat^ GetOutputFormat(String^ shortName)
		{
			return GuessType(shortName, "", "");
		}

		///<summary>Guesses a video format from some given information.</summary>
		///<remarks>Calls guess_format.</remarks>
		static LibAvOutputFormat^ GuessType(String^ shortName, String^ filename, String^ mimeType);

		///<summary>Returns the memory footprint for width x height image.</summary>
		///<remarks>Calls avpicture_get_size.</remarks>
		static int PictureGetSize(LibAvWrap::PixelFormat pixelFormat, int width, int height)
		{
			return avpicture_get_size((::PixelFormat)pixelFormat, width, height);
		}

		///<summary>Initialises libavformat and register all the muxers, demuxers and protocols.</summary>
		///<remarks>Calls av_register_all.</remarks>
		static void RegisterAll() 
		{
			av_register_all();
		}

		///<summary>Error handling helper function: throws an exception if the result is not zero.</summary>
		static void ThrowIfError(int result)
		{
			// check if everything is okay
			if(result == 0)
				return;

			// otherwise look for the error
			switch(result)
			{
				case -2:  // AVERROR(ENOENT) = AVERROR_NOENT
					throw gcnew System::IO::FileNotFoundException("No such file or directory.");
				case -5:  // AVERROR(EIO) = AVERROR_IO
					throw gcnew Exception("I/O error occured (usually that means that input file is truncated and/or corrupted).");
				case -12: // AVERROR(ENOMEM) = AVERROR_NOMEM
					throw gcnew Exception("Memory allocation error occured.");
				case -22: // AVERROR(EINVAL) = AVERROR_UNKNOWN or AVERROR_INVALIDDATA
					throw gcnew Exception("Error while parsing header.");
				case -32: // AVERROR(EPIPE) = AVERROR_EOF
					throw gcnew Exception("End of file or broken pipe.");
				case -33: // AVERROR(EDOM) = AVERROR_NUMEXPECTED
					throw gcnew Exception("Numerical argument out of domain.");
				case -78: // AVERROR(ENOSYS) = AVERROR_NOTSUPP
					throw gcnew Exception("Function not supported.");
				case -92: // AVERROR(EILSEQ) = AVERROR_NOFMT
					throw gcnew Exception("Unknown format.");
				default:
					throw gcnew Exception(String::Format("Error {0} while opening file.", result));
			}
		}

		///<summary>Error handling helper function: throws an exception if the result is not zero.</summary>
		static void ThrowIfError(String^ filename, int result)
		{
			// check if everything is okay
			if(result == 0)
				return;

			// otherwise look for the error
			switch(result)
			{
				case -2:  // AVERROR(ENOENT) = AVERROR_NOENT
					throw gcnew System::IO::FileNotFoundException(filename + ": no such file or directory.", filename);
				case -5:  // AVERROR(EIO) = AVERROR_IO
					throw gcnew Exception(filename + ": I/O error occured (usually that means that input file is truncated and/or corrupted).");
				case -12: // AVERROR(ENOMEM) = AVERROR_NOMEM
					throw gcnew Exception(filename + ": memory allocation error occured.");
				case -22: // AVERROR(EINVAL) = AVERROR_UNKNOWN or AVERROR_INVALIDDATA
					throw gcnew Exception(filename + ": Error while parsing header.");
				case -32: // AVERROR(EPIPE) = AVERROR_EOF
					throw gcnew Exception(filename + ": end of file or broken pipe.");
				case -33: // AVERROR(EDOM) = AVERROR_NUMEXPECTED
					throw gcnew Exception(filename + ": numerical argument out of domain.");
				case -78: // AVERROR(ENOSYS) = AVERROR_NOTSUPP
					throw gcnew Exception(filename + ": function not supported.");
				case -92: // AVERROR(EILSEQ) = AVERROR_NOFMT
					throw gcnew Exception(filename + ": Unknown format.");
				default:
					throw gcnew Exception(String::Format("{0}: Error {1} while opening file.", filename, result));
			}
		}

		///<summary>Close the ByteIOContext associated with a file or URL.</summary>
		///<remarks>Calls url_fclose.</remarks>
		static void UrlFClose(LibAvByteIOContext^ bioc);

		///<summary>Opens a file or URL.</summary>
		///<remarks>Calls url_fopen().</remarks>
		//static int UrlFOpen(AvDnByteIOContext^ bioc, String^ filename, int flag); // FIXME: doesn't work
		static int UrlFOpen(LibAvFormatContext^ fc, String^ filename, int flag);

		///<summary>Install a printf-based logger.</summary>
		static void UsePrintfLogger()
		{
			av_log_set_callback(&printf_callback);
		}

// properties -------------------------------------------------------------------------------------

		///<summary>Log level of the libav* libraries.</summary>
		static property int LogLevel {
			int get() { return av_log_get_level(); }
			void set(int level) { av_log_set_level(level); }
		}
	};

} // namespace LibAvWrap