﻿// $Id: FlowVisualiser.cs 460 2010-10-25 12:11:31Z cr333 $
using System;
using System.ComponentModel;
using System.Drawing;
using OpenTK;
using OpenTK.Graphics.OpenGL;
using VideoLib.Frames;

namespace VideoLib.OpenGL.OpticalFlow
{
	/// <summary>Visualises optical flow using arrows in a grid.</summary>
	public class FlowVisualiser : Node, IOpenGLNode
	{
		/// <summary>Gets or sets the width of the output frame.</summary>
		[Description("Width of the frame to render.")]
		public int Width  { get; set; }

		/// <summary>Gets or sets the height of the output frame.</summary>
		[Description("Height of the frame to render.")]
		public int Height { get; set; }

		/// <summary>Gets or sets the grid distance between arrows.</summary>
		[DefaultValue(8), Description("The distance between arrows, in flow pixels.")]
		public int GridSize { get; set; }

		/// <summary>Gets or sets the color of the flow arrows.</summary>
		[Description("Colour of the arrows drawn.")]
		public Color ArrowColor { get; set; }

		/// <summary>Gets or sets a factor by which to scale the arrows. Default: 1.</summary>
		[DefaultValue(1.0f), Description("Scale the arrows by this factor.")]
		public float ArrowScale { get; set; }

		/// <summary>Initialises a new instance of the <see cref="FlowVisualiser"/> class.</summary>
		public FlowVisualiser() : base()
		{
			// default settings
			Width  = 400;
			Height = 300;
			GridSize = 8;
			ArrowColor = Color.LightYellow;
			ArrowScale = 1;

			AddInputPins("flow", "image");
			AddOutputPin("out");
		}

		/// <summary>Processes some frames, to produce new frames.</summary>
		public override void Process()
		{
			// get the input textures
			var frames = GetInputFrames();
			if (PassOnEndOfStreamFrame(frames)) return;
			var textures = CastFramesTo<TextureFrame>(frames);

			// read out flow texture (need access to pixel values)
			var flow = textures["flow"].Texture.ReadFrame(VideoLib.Frames.PixelFormat.FloatBGRA);

			// prepare a new texture frame to render into
			var result = new TextureFrame(Width, Height)
			{
				InternalFormat = PixelInternalFormat.Rgb,
				SequenceNumber = textures["image"].SequenceNumber
			};

			// rendering
			SetupFBO(result.Texture);
			InitDrawing();
			DrawImage(textures["image"]);
			DrawArrows(flow);
			StopDrawing();

			PushFrame(result);

			// clean up
			textures["flow"].RefDec();
			textures["image"].RefDec();
			result.RefDec();
		}

		#region Drawing functions
//=================================================================================================

		private int FramebufferID;

		private void SetupFBO(Texture renderTexture)
		{
			// ensure correct size of render texture
			renderTexture.ReAlloc(Width, Height);

			// set up FBO for drawing
			if (FramebufferID == 0) GL.GenFramebuffers(1, out FramebufferID);
			GL.BindFramebuffer(FramebufferTarget.DrawFramebuffer, FramebufferID);
			GL.FramebufferTexture2D(FramebufferTarget.DrawFramebuffer, FramebufferAttachment.ColorAttachment0, TextureTarget.Texture2D, renderTexture, 0);
			GL.DrawBuffer(DrawBufferMode.ColorAttachment0);
			Utils.CheckFboStatus();
		}

		private void InitDrawing()
		{
			GL.BindFramebuffer(FramebufferTarget.DrawFramebuffer, FramebufferID);
			GL.Viewport(0, 0, Width, Height);

			// use blending for anti-aliased line drawing
			GL.Enable(EnableCap.Blend);
			GL.BlendFunc(BlendingFactorSrc.SrcAlpha, BlendingFactorDest.OneMinusSrcAlpha);

			GL.Enable(EnableCap.Texture2D);
			GL.Disable(EnableCap.DepthTest);
			GL.DepthMask(false);

			GL.MatrixMode(MatrixMode.Projection);
			GL.LoadIdentity();

			GL.MatrixMode(MatrixMode.Modelview);
			GL.LoadIdentity();
		}

		private void DrawImage(TextureFrame image)
		{
			image.Bind(TextureUnit.Texture0);
			GL.Begin(BeginMode.Quads);
			GL.TexCoord2(0, 0); GL.Vertex2(-1, -1);
			GL.TexCoord2(1, 0); GL.Vertex2( 1, -1);
			GL.TexCoord2(1, 1); GL.Vertex2( 1,  1);
			GL.TexCoord2(0, 1); GL.Vertex2(-1,  1);
			GL.End();
			image.Unbind();
		}

		private unsafe void DrawArrows(BitmapFrame flowFrame)
		{
			fixed (byte* flowBytes = &flowFrame.Data[0])
			{
				Vector4* flows = (Vector4*)(void*)flowBytes;

				GL.Enable(EnableCap.LineSmooth);
				GL.Begin(BeginMode.Lines);
				GL.Color3(ArrowColor);
				for (int y = GridSize / 2; y < flowFrame.Height; y += GridSize)
				{
					for (int x = GridSize / 2; x < flowFrame.Width; x += GridSize)
					{
						Vector4 flow = flows[y * flowFrame.Width + x];

						// compute start and end points of arrow
						var start = new Vector2(2.0f * (x                      ) / flowFrame.Width - 1, 2.0f * (y                       ) / flowFrame.Height - 1);
						var end   = new Vector2(2.0f * (x - flow.X * ArrowScale) / flowFrame.Width - 1, 2.0f * (y - flow.Y  * ArrowScale) / flowFrame.Height - 1);

						// draw line segment
						GL.Vertex2(start); GL.Vertex2(end);

						// don't draw arrow head if too short
						var length = Vector2.Subtract(start, end).LengthFast;
						if (length * Width / 2 < 4) continue;

						// angle of arrow line to x-axis
						var angle = Math.Atan2(flow.Y, flow.X);

						// scale size of arrow head with length (within limit)
						length = Math.Min(length, 2 * 20 / (float)Width);

						// compute arrow points
						var left  = new Vector2(end.X + length / 3 * (float)Math.Cos(angle + Math.PI / 8), end.Y + length / 3 * (float)Math.Sin(angle + Math.PI / 8));
						var right = new Vector2(end.X + length / 3 * (float)Math.Cos(angle - Math.PI / 8), end.Y + length / 3 * (float)Math.Sin(angle - Math.PI / 8));

						// draw arrow head lines
						GL.Vertex2(end); GL.Vertex2(left);
						GL.Vertex2(end); GL.Vertex2(right);
					}
				}

				GL.Color3(Color.White);
				GL.End();
				GL.Disable(EnableCap.LineSmooth);
			}
		}

		private void StopDrawing()
		{
			GL.BindFramebuffer(FramebufferTarget.DrawFramebuffer, 0);
			GL.Enable(EnableCap.DepthTest);
			GL.DepthMask(true);
			GL.Disable(EnableCap.Blend);
		}

//=================================================================================================
		#endregion
	}
}