﻿// $Id: TraceVisualiser.cs 375 2010-08-30 19:18:05Z cr333 $
using System;
using System.Drawing;
using OpenTK;
using OpenTK.Graphics.OpenGL;
using VideoLib.Frames;

namespace VideoLib.OpenGL.OpticalFlow
{
	/// <summary>Visualises the optical flow using traced lines.</summary>
	public class TraceVisualiser : Node, IOpenGLNode
	{
		/// <summary>Gets or sets the width of the output frame.</summary>
		public int Width  { get; set; }

		/// <summary>Gets or sets the height of the output frame.</summary>
		public int Height { get; set; }

		/// <summary>Gets or sets the grid distance between arrows.</summary>
		public int GridSize { get; set; }

		/// <summary>Gets or sets the color of the flow arrows.</summary>
		public Color LineColor { get; set; }

		/// <summary>Gets or sets whether the current image should be drawn. Default: false.</summary>
		/// <remarks><c>true</c> shows the 'particle' mode; <c>false</c> shows the 'trace' mode.</remarks>
		public bool ShowImage { get; set; }

		/// <summary>Gets or sets the fade factor (0..1). The higher, the faster the fading.</summary>
		public float FadeFactor { get; set; }

		/// <summary>Initialises a new instance of the <see cref="TraceVisualiser"/> class.</summary>
		public TraceVisualiser() : base()
		{
			// default settings
			Width      = 400;
			Height     = 300;
			GridSize   = 8;
			LineColor  = Color.LightYellow;
			ShowImage  = false;
			FadeFactor = 0.1f;

			AddInputPins("image", "flow");
			AddOutputPin("out");
		}

		/// <summary>Processes some frames, to produce new frames.</summary>
		public override void Process()
		{
			// get the input bitmaps
			var frames = GetInputFrames();
			if (PassOnEndOfStreamFrame(frames)) return;
			var bitmaps = CastFramesTo<BitmapFrame>(frames);

			// check format of flow
			if ((bitmaps["flow"].Format & VideoLib.Frames.PixelFormat.TypeMask) != VideoLib.Frames.PixelFormat.Single)
				throw new NotSupportedException(String.Format("Flow needs to have float components (format is {0}).", bitmaps["flow"].Format));
			if ((bitmaps["flow"].Format & VideoLib.Frames.PixelFormat.FourColours) == VideoLib.Frames.PixelFormat.None)
				throw new NotSupportedException(String.Format("Flow needs to have 4 components (format is {0}).", bitmaps["flow"].Format));

			// set up texture
			if (ImageTexture == null)
				ImageTexture = new Texture(bitmaps["image"]) { InternalFormat = PixelInternalFormat.Rgb };

			// rendering
			SetupFBO();
			InitDrawing();
			if(ShowImage) DrawImage(bitmaps["image"]);
			else          DrawTransparentQuad();
			DrawLines(bitmaps["flow"]);
			StopDrawing();

			// read back render texture
			var result = RenderTexture.ReadFrame(VideoLib.Frames.PixelFormat.BGR24);
			result.SequenceNumber = bitmaps["image"].SequenceNumber;
			PushFrame(result);
		}

		#region Drawing functions
//=================================================================================================

		private Texture ImageTexture;
		private Vector2[] Positions;
		private int FramebufferID;
		private Texture RenderTexture;
		private Size FramebufferSize;

		private void SetupFBO()
		{
			// correct size already? nothing to do then ...
			if (Width == FramebufferSize.Width && Height == FramebufferSize.Height) return;

			// set up rendering texture
			if (RenderTexture == null) RenderTexture = new Texture(Width, Height) { InternalFormat = PixelInternalFormat.Rgb };
			RenderTexture.ReAlloc(Width, Height);

			// set up FBO for drawing
			if (FramebufferID == 0) GL.GenFramebuffers(1, out FramebufferID);
			GL.BindFramebuffer(FramebufferTarget.DrawFramebuffer, FramebufferID);
			GL.FramebufferTexture2D(FramebufferTarget.DrawFramebuffer, FramebufferAttachment.ColorAttachment0, TextureTarget.Texture2D, RenderTexture.ID, 0);
			GL.DrawBuffer(DrawBufferMode.ColorAttachment0);
			Utils.CheckFboStatus();

			FramebufferSize = new Size(Width, Height);
		}

		private void InitDrawing()
		{
			GL.BindFramebuffer(FramebufferTarget.DrawFramebuffer, FramebufferID);
			GL.Viewport(0, 0, FramebufferSize.Width, FramebufferSize.Height);

			// use blending for anti-aliased line drawing
			GL.Enable(EnableCap.Blend);
			GL.BlendFunc(BlendingFactorSrc.SrcAlpha, BlendingFactorDest.OneMinusSrcAlpha);

			GL.Enable(EnableCap.Texture2D);
			GL.Disable(EnableCap.DepthTest);
			GL.DepthMask(false);

			GL.MatrixMode(MatrixMode.Projection);
			GL.LoadIdentity();

			GL.MatrixMode(MatrixMode.Modelview);
			GL.LoadIdentity();
		}

		private void DrawImage(BitmapFrame image)
		{
			ImageTexture.Update(image);
			ImageTexture.Bind();
			GL.Begin(BeginMode.Quads);
			GL.TexCoord2(0, 0); GL.Vertex2(-1, -1);
			GL.TexCoord2(1, 0); GL.Vertex2( 1, -1);
			GL.TexCoord2(1, 1); GL.Vertex2( 1,  1);
			GL.TexCoord2(0, 1); GL.Vertex2(-1,  1);
			GL.End();
			ImageTexture.Unbind();
		}

		private void DrawTransparentQuad()
		{
			GL.Begin(BeginMode.Quads);
			GL.Color4(0, 0, 0, FadeFactor);
			GL.Vertex2(-1, -1);
			GL.Vertex2( 1, -1);
			GL.Vertex2( 1,  1);
			GL.Vertex2(-1,  1);
			GL.End();
		}

		private unsafe void DrawLines(BitmapFrame flowFrame)
		{
			// initialise positions using regular grid
			if (Positions == null)
			{
				int xPoints = (flowFrame.Width  - GridSize / 2) / GridSize;
				int yPoints = (flowFrame.Height - GridSize / 2) / GridSize;
				Positions = new Vector2[(xPoints + 1) * (yPoints + 1)];

				int index = 0;
				for (int y = GridSize / 2; y < flowFrame.Height; y += GridSize)
				{
					for (int x = GridSize / 2; x < flowFrame.Width; x += GridSize)
					{
						Positions[index++] = new Vector2(x, y);
					}
				}
			}

			fixed (byte* flowBytes = &flowFrame.Data[0])
			{
				Vector4* flows = (Vector4*)(void*)flowBytes;

				GL.Enable(EnableCap.LineSmooth);
				GL.Begin(BeginMode.Lines);
				GL.Color3(LineColor);

				for(int i = 0; i < Positions.Length; i++)
				{
					// round to nearest pixel
					int x = (int)(Positions[i].X + 0.5f);
					int y = (int)(Positions[i].Y + 0.5f);

					// ignore outliers
					if (x < 0 || x >= flowFrame.Width ) continue;
					if (y < 0 || y >= flowFrame.Height) continue;

					// look up the flow at this pixel
					Vector4 flow = flows[y * flowFrame.Width + x];

					// starting point of line segment
					var start = new Vector2(2 * Positions[i].X / flowFrame.Width - 1, 2 * Positions[i].Y / flowFrame.Height - 1);

					// update position using optical flow
					Positions[i].X += flow.Z;
					Positions[i].Y += flow.W;

					// end points of line segment
					var end = new Vector2(2 * Positions[i].X / flowFrame.Width - 1, 2 * Positions[i].Y / flowFrame.Height - 1);

					// draw line segment
					GL.Vertex2(start); GL.Vertex2(end);					
				}

				GL.Color3(Color.White);
				GL.End();
				GL.Disable(EnableCap.LineSmooth);
			}
		}

		private void StopDrawing()
		{
			GL.BindFramebuffer(FramebufferTarget.DrawFramebuffer, 0);
			GL.Enable(EnableCap.DepthTest);
			GL.DepthMask(true);
			GL.Disable(EnableCap.Blend);
		}

//=================================================================================================
		#endregion
	}
}