﻿// $Id: UniformInfo.cs 514 2010-12-17 12:48:14Z cr333 $
using System;
using System.Reflection;
using OpenTK;
using OpenTK.Graphics.OpenGL;

namespace VideoLib.OpenGL
{
	/// <summary>
	/// Represents a set of details about a uniform (a CLR property marked with [<see cref="UniformAttribute"/>]),
	/// obtained using reflection on a certain type.
	/// </summary>
	public class UniformInfo
	{
		/// <summary>Initialises a new instance of the <see cref="UniformInfo"/> class.</summary>
		/// <param name="instance">Class instance for access to property value.</param>
		/// <param name="propertyInfo">Information about the class property.</param>
		/// <param name="uniformName">Name of the uniform in the shader program.</param>
		/// <param name="location">The location of the uniform in the shader program.</param>
		/// <param name="attribute">The uniform attribute.</param>
		public UniformInfo(object instance, PropertyInfo propertyInfo, string uniformName, int location, UniformAttribute attribute)
		{
			Instance  = instance;
			Property  = propertyInfo;
			Name      = uniformName;
			Location  = location;
			Attribute = attribute;
		}

		/// <summary>Gets or sets the class instance for access to property value.</summary>
		public Object Instance { get; protected set; }

		/// <summary>Gets or sets information about the class property.</summary>
		public PropertyInfo Property { get; protected set; }

		/// <summary>Gets or sets the name of the uniform in the shader program.</summary>
		public string Name { get; protected set; }

		/// <summary>Gets or sets the location of the uniform in the shader program.</summary>
		public int Location { get; protected set; }

		/// <summary>Gets or sets the attribute for this uniform.</summary>
		public UniformAttribute Attribute { get; set; }

		/// <summary>Stores the most recent value the uniform was set to.</summary>
		private object lastValue;

		/// <summary>Updates the uniform using the current property value (only if it changes).</summary>
		public void UpdateUniform()
		{
			// check for valid location
			if (Location < -1) return;
			if (Location == -1)
			{
				Console.WriteLine(String.Format("Warning: The uniform '{0}' could not be found in {1}'s shader program.", Name, Instance));
				Location--; // to only show warning once
				return;
			}

			var value = Property.GetValue(Instance, null);
			var type  = Property.PropertyType;

			// always bind textures
			if (Property.PropertyType.Equals(typeof(Texture)))
			{
				var texture = (Texture)value;
				texture.Bind(TextureUnit.Texture0 + Attribute.Unit);
				return;
			}

			// check if value has changed; skip update if not
			if (lastValue != null && lastValue.Equals(value))
				return;

			// TODO: arrays are not recognised as changing, as Equals doesn't look at element values
			//       (would need a deep copy of array when storing lastValue, and array comparison above)

			if      (type.Equals(typeof(float)))   GL.Uniform1(Location,   (float)value);
			else if (type.Equals(typeof(uint)))    GL.Uniform1(Location,    (uint)value);
			else if (type.Equals(typeof(int)))     GL.Uniform1(Location,     (int)value);
			else if (type.Equals(typeof(bool)))    GL.Uniform1(Location,    (bool)value ? 1 : 0); // pass bool as int
			else if (type.IsEnum)                  GL.Uniform1(Location,     (int)value); // pass enum as int
			else if (type.Equals(typeof(Vector2))) GL.Uniform2(Location, (Vector2)value);
			else if (type.Equals(typeof(Vector3))) GL.Uniform3(Location, (Vector3)value);
			else if (type.Equals(typeof(Vector4))) GL.Uniform4(Location, (Vector4)value);
			else if (type.Equals(typeof(float[])))   { var vector = (float[])value;   GL.Uniform1(Location, vector.Length, ref vector[0]); }
			else if (type.Equals(typeof(Vector2[]))) { var vector = (Vector2[])value; GL.Uniform2(Location, vector.Length, ref vector[0].X); }
			else if (type.Equals(typeof(Vector3[]))) { var vector = (Vector3[])value; GL.Uniform3(Location, vector.Length, ref vector[0].X); }
			else if (type.Equals(typeof(Vector4[]))) { var vector = (Vector4[])value; GL.Uniform4(Location, vector.Length, ref vector[0].X); }
			else throw new NotImplementedException(String.Format("Cannot handle type '{0}'.", type));

			// retain value for next call
			lastValue = value;
		}

		/// <summary>Returns a <see cref="System.String"/> that represents this instance.</summary>
		/// <returns>A <see cref="System.String"/> that represents this instance.</returns>
		public override string ToString()
		{
			return string.Format("{0} {1} => {2}", Property.PropertyType.Name, Name, Location);
		}
	}
}
