﻿// $Id: UniformReflector.cs 506 2010-11-26 19:18:42Z cr333 $
using System.Collections.Generic;
using System.Reflection;
using OpenTK.Graphics.OpenGL;

namespace VideoLib.OpenGL
{
	/// <summary>Contains static methods to reflect on properties marked with [<see cref="UniformAttribute"/>].</summary>
	public static class UniformReflector
	{
		/// <summary>Enumerates all marked uniforms in the given class instance.</summary>
		/// <param name="instance">Class instance containing the uniforms.</param>
		/// <param name="program">The shader program containing the uniforms.</param>
		/// <returns>A list of all found uniforms.</returns>
		public static IEnumerable<UniformInfo> EnumerateUniforms(object instance, int program)
		{
			List<UniformInfo> uniforms = new List<UniformInfo>();

			// first enumerate all properties
			var properties = instance.GetType().GetProperties(BindingFlags.Instance | BindingFlags.Public | BindingFlags.NonPublic);
			foreach (var property in properties)
			{
				// only consider the ones marked with UniformAttribute
				var attr = property.GetCustomAttributes(typeof(UniformAttribute), false);
				if (attr != null && attr.Length > 0)
				{
					var uniform = (UniformAttribute)attr[0];

					// If the name is not set explicitly, use the camelCase version of the PascalCase property name.
					// For example, if the property name is "LineWidth", the uniform name will be "lineWidth".
					var name = uniform.Name;
					if (name == "") name = property.Name.Substring(0, 1).ToLower() + property.Name.Substring(1);

					// look up uniform location once and reuse later on all updates
					var location = GL.GetUniformLocation(program, name);

					uniforms.Add(new UniformInfo(instance, property, name, location, uniform));
				}
			}

			return uniforms;
		}
	}
}
