﻿// $Id: OutputPin.cs 436 2010-10-15 16:26:38Z cr333 $
using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using VideoLib.Frames;

namespace VideoLib
{
	/// <summary>Class encapsulating a <see cref="Node"/>'s output pin.</summary>
	public class OutputPin : Pin
	{
		/// <summary>Current status of this output pin.</summary>
		public override PinStatus Status
		{
			// getting status is always up-to-date
			get { return Node.Topology.UpdateStatus(this); }

			// setting the status of the parent (Pin), raises StatusChanged event
			protected set { base.Status = value; }
		}

		/// <summary>Current status of this output Pin.</summary>
		public OutputPin(string name, INode node) : base(name, node)
		{
			Status = PinStatus.Ready;
		}

		/// <summary>Pushes a frame to any listeners to <see cref="Pin.NewFrame"/>.</summary>
		/// <param name="frame">The new frame.</param>
		public void PushFrame(Frame frame)
		{
			// invoke a NewFrame event, to signal to all registered input pins
			OnNewFrame(frame);

			// As downstream input pins might have become blocked, we need to update
			// this output pin's status. This in turn might change the status of its node.
			//
			// So we should do:
			//   1. UpdateStatus();
			//   2. Node.UpdateStatus();
			//
			// However, this is not necessary, as this PushFrame method should only be
			// called by a running node which will update its own status when it's done.
		}

		/// <summary>Request the pin to update its status.</summary>
		public override void UpdateStatus()
		{
			Status = Status;
		}

		/// <summary>Returns a string representation of this OutputPin.</summary>
		/// <returns>For example "LibAvCodecSource.out(Ready)" for pin "out".</returns>
		public override string ToString()
		{
			return String.Format("{0}({1})", base.ToString(), Status);
		}

		/// <summary>Listens for downstream events that might block this pin.</summary>
		/// <param name="sender">The sender of the event (usually an <see cref="VideoLib.InputPin"/>.</param>
		/// <param name="e">The <see cref="VideoLib.PinStatusChangedEventArgs"/> instance containing the event data.</param>
		public void DownStreamStatusChangedListener(object sender, PinStatusChangedEventArgs e)
		{
			// if anything downstream becomes blocked, we are blocked
			if (e.NewStatus == PinStatus.Blocked)
				Status = PinStatus.Blocked;
			else // need to update status
				UpdateStatus();
		}
	}
}
