﻿// $Id: ModularScheduler.cs 440 2010-10-18 11:49:09Z cr333 $
// Based on a Part II project by Rubin Xu (2010).
// Derived from SimpleDequeScheduler.cs
using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Threading;
using VideoLib.Frames;

namespace VideoLib.Scheduling
{
	/// <summary>Implements a modular scheduler supporting several work queues.</summary>
	/// <remarks>
	/// This scheduler maintains a configurable number of work queues, and can be set up as required.
	/// The work queue implementations are provided at construction time. The order in which they are
	/// given defines the order in which events are handled. Once an event has been handled by a work
	/// queue, the event will be withheld from the remaining queues that follow.
	/// 
	/// IMPORTANT: This scheduler, as well as <see cref="CudaAwareScheduler"/> and
	/// <see cref="SimpleDequeScheduler"/> are the only working schedulers at the moment,
	/// due to changes in how events are handled.
	/// </remarks>
	public class ModularScheduler : Scheduler
	{
		/// <summary>Constructs a modular queue-based scheduler.</summary>
		/// <param name="topology">Topology of nodes to use.</param>
		/// <param name="queues">Sequence of work queues, in the order they will be processed in.</param>
		public ModularScheduler(Topology topology, params WorkQueue[] queues) : base(topology)
		{
			Queues = new List<WorkQueue>(queues);
		}

		/// <summary>Constructs a modular queue-based scheduler with one worker thread per processor.</summary>
		/// <param name="topology">Topology of nodes to use.</param>
		public ModularScheduler(Topology topology)
			: this(topology, new SharedWorkQueue(Environment.ProcessorCount))
		{
		}

		/// <summary>List of work queues that handle events in order.</summary>
		public List<WorkQueue> Queues { get; set; }

		/// <summary>Sets up the scheduler, preparing it for the work ahead.</summary>
		public override void Setup()
		{
			base.Setup();
			try
			{
				// initialise status
				Status = SchedulerStatus.Starting;

				// set up work queues
				foreach (var queue in Queues)
				{
					queue.Scheduler = this;
					queue.Topology = Topology;
				}

				// listen to all new frames and status updates from output pins
				foreach (var edgeFan in Topology.OutgoingEdges)
				{
					edgeFan.Key.NewFrame += NewFrameListener;
					edgeFan.Key.StatusChanged += PinStatusChangedListener;
				}

				// listen to all status updates from input pins
				foreach (var edgeFan in Topology.IncomingEdges)
					edgeFan.Key.StatusChanged += PinStatusChangedListener;

				// listen to all status changes from nodes
				foreach (INode node in Topology.Nodes)
					node.StatusChanged += NodeStatusChangedListener;

				// let's go over to proper running mode
				Status = SchedulerStatus.Running;
			}
			catch (Exception)
			{
				throw;
			}
		}

		/// <summary>Starts the worker threads associated with the work queues.</summary>
		/// <param name="numWorkers">Ignored.</param>
		public override void StartWorkers(int numWorkers)
		{
			foreach (var module in Queues)
				workers.AddRange(module.StartWorkers());

			// ask each queue to initialise nodes, stop after it was handled
			foreach (var node in Topology.Nodes)
				foreach (var queue in Queues)
					if (queue.HandleNodeInitEvent(node))
						break;
		}

		/// <summary>Dummy main function of a worker thread.</summary>
		protected override void RunWorker(object data)
		{
			throw new NotImplementedException();
		}

		/// <summary>Listens to new frames and schedules transport of new frames to receivers.</summary>
		/// <param name="sender">The sender.</param>
		/// <param name="e">The <see cref="VideoLib.Frames.NewFrameEventArgs"/> instance containing the event data.</param>
		private void NewFrameListener(object sender, NewFrameEventArgs e)
		{
			// ask each queue to handle the event, stop after it was handled
			foreach (var queue in Queues)
				if (queue.HandleNewFrameEvent(sender, e))
					break;
		}

		/// <summary>Listens to status changes in IO pins.</summary>
		/// <param name="sender">The sender.</param>
		/// <param name="e">The <see cref="VideoLib.PinStatusChangedEventArgs"/> instance containing the event data.</param>
		private void PinStatusChangedListener(object sender, PinStatusChangedEventArgs e)
		{
			// ask each queue to handle the event, stop after it was handled
			foreach (var queue in Queues)
				if (queue.HandlePinStatusChangedEvent(sender, e))
			        break;
		}

		/// <summary>Listens to status changes of nodes.</summary>
		/// <param name="sender">The sender.</param>
		/// <param name="e">The <see cref="VideoLib.NodeStatusChangedEventArgs"/> instance containing the event data.</param>
		private void NodeStatusChangedListener(object sender, NodeStatusChangedEventArgs e)
		{
			Debug.WriteLine(String.Format("{0}: {1} changed: {2} => {3}", Thread.CurrentThread.Name, sender, e.OldStatus, e.NewStatus));
			INode node = (INode)sender;

			// if a node is ready to run, append it to some ready queue
			if (e.NewStatus == NodeStatus.Ready)
			{
				// ask each queue to handle the event, stop after it was handled
				foreach (var module in Queues)
					if (module.HandleNodeReadyEvent(node))
						break;

				return;
			}

			// if a node is now waiting for data, there is nothing we need to do
			if (e.NewStatus == NodeStatus.Waiting) return;

			// if a node is was ready and is now running, there is nothing we need to do
			if (e.OldStatus == NodeStatus.Ready && e.NewStatus == NodeStatus.Running) return;

			// if a node is finished with processing, there is nothing we need to do
			if (e.NewStatus == NodeStatus.Finished) return;

			// if a node is now stopped, exit everything
			if (e.NewStatus == NodeStatus.Stopped)
			{
				Console.Error.WriteLine("Node " + node.Name + " was stopped.");
				Stop();
				return;
			}

			throw new NotImplementedException(String.Format("Cannot handle state transition {0} => {1} of node {2}.", e.OldStatus, e.NewStatus, node));
		}
	}
}
