﻿// $Id: RawSource.cs 494 2010-11-15 20:31:09Z cr333 $
using System;
using System.IO;
using System.Threading;
using VideoLib.Frames;
using System.Diagnostics;

namespace VideoLib.Sources
{
	/// <summary>Loads a sequence of raw bitmap frames from disk.</summary>
	public class RawSource : Node
	{
		/// <summary>Filename template for use with String.Format, e.g. "{0:D4}.png" for "NNNN.png".</summary>
		public string Filenames { get; protected set; }

		/// <summary>The sequence number of the most recently loaded image.</summary>
		public int SequenceNumber { get; protected set; }

		/// <summary>The framerate with which the video frames will be output.</summary>
		public float FPS { get; set; }

		/// <summary>Sequence number of the first frame. Default: 0.</summary>
		public int FirstNumber { get; set; }

		/// <summary>Sequence number of the last frame. Default: -1 (infinite).</summary>
		public int LastNumber { get; set; }

		/// <summary>Restart sequence after the last frame. Default: false.</summary>
		public bool Loop { get; set; }

		/// <summary>Human-friendly node name.</summary>
		public override string Name
		{
			get { return String.Format("{0}({1})", base.Name, Path.GetFileName(String.Format(Filenames, 1234))); }
		}

		/// <summary>Maximum number of missing images in the image sequence. Default: 0 (no missing images).</summary>
		public int MaxGap { get; set; }

		/// <summary>Last frame that was loaded successfully.</summary>
		private int lastFrameNumber;

		/// <summary>Stopwatch time of the last frame, to control FPS.</summary>
		private long lastFrameTime;

		/// <summary>This stopwatch is used to control the frame rate.</summary>
		private Stopwatch stopwatch = Stopwatch.StartNew();

		/// <summary>Constructs a new frame sequence source.</summary>
		/// <param name="filenames">Template for filenames for use with String.Format, e.g. "{0:D4}.png" for "NNNN.png".</param>
		public RawSource(string filenames)
			: base()
		{
			// Default settings
			Filenames = filenames;
			FirstNumber = 0;
			LastNumber = -1;
			MaxGap = 0;
			FPS = 25;
			Loop = false;

			AddOutputPin("out");
		}

		/// <summary>Prepares the image sequence source.</summary>
		public override void Init()
		{
			SequenceNumber = FirstNumber;
		}

		/// <summary>Processes a frame.</summary>
		public override void Process()
		{
			// read in the next image, skipping missing ones if required
			do
			{
				// check if we're still in the valid range of sequence numbers
				if (LastNumber != -1 && SequenceNumber > LastNumber)
				{
					if (Loop)
					{
						SequenceNumber = FirstNumber;
					}
					else
					{
						PushFrame("out", new EndOfStreamFrame() { SequenceNumber = SequenceNumber });
						Status = NodeStatus.Finished;
						return;
					}
				}

				// filename of the current image
				string filename = String.Format(Filenames, SequenceNumber);

				if (File.Exists(filename))
				{
					// load image and set sequence number
					BitmapFrame frame = LoadFrameFromFile(filename);
					frame.SequenceNumber = SequenceNumber;

					// push new frame, updating internal state
					PushFrame("out", frame);
					lastFrameNumber = SequenceNumber;
					SequenceNumber++;

					// sleep to get framerate right
					float sleepTime = 1000 / FPS - (stopwatch.ElapsedMilliseconds - lastFrameTime);
					Thread.Sleep(Math.Max(0, (int)sleepTime));
					lastFrameTime = stopwatch.ElapsedMilliseconds;
					return;
				}
				else
				{
					// skip missing image
					SequenceNumber++;
					//continue;
				}

				// check if the gap has been too long
				if (SequenceNumber > lastFrameNumber + MaxGap + 1)
				{
					if (Loop)
					{
						SequenceNumber = FirstNumber;
					}
					else
					{
						PushFrame("out", new EndOfStreamFrame() { SequenceNumber = SequenceNumber });
						Status = NodeStatus.Finished;
						return;
					}
				}
			} while (SequenceNumber <= lastFrameNumber + MaxGap + 1);
		}

		/// <summary>Loads a bitmap frame from a file.</summary>
		/// <param name="filename">The filename.</param>
		public static BitmapFrame LoadFrameFromFile(string filename)
		{
			using (var stream = new FileStream(filename, FileMode.Open, FileAccess.Read))
			{
				using (var reader = new BinaryReader(stream))
				{
					string header = reader.ReadString();
					if (header != "BitmapFrame")
						throw new NotSupportedException(String.Format("Frame type \"{0}\" not supported.", header));

					uint version = reader.ReadUInt32();
					if (version != 1u)
						throw new NotSupportedException(String.Format("Raw format version \"{0}\" not supported.", version));

					int width = reader.ReadInt32();
					int height = reader.ReadInt32();
					var format = (PixelFormat)reader.ReadUInt32();
					int stride = reader.ReadInt32();

					var frame = new BitmapFrame(width, stride, height, format);
					int read = reader.Read(frame.Data, 0, stride * height);
					if (read != stride * height)
						throw new NotSupportedException(String.Format("File too short (need {0} bytes, read {1}).", stride * height, read));

					return frame;
				}
			}
		}
	}
}
